/*                                    */
/* foobar2000 wrapper for SNESAPU.dll */
/*                                    */
/*        hereby public domain        */
/*                                    */

#include "Types.h"
#include "SNESAPU.h"
#include "resource.h"
#include "../SDK/foobar2000.h"
#include "script700_parser.h"

#define SPC_ID "SNES-SPC700 Sound File Data "
#define XID6_ID "xid6"
#define OFS_XID6 0x10200

//Sub-chunk ID's (from SNES AMP)
#define	XID6_SONG	0x01
#define	XID6_GAME	0x02
#define	XID6_ARTIST	0x03
#define	XID6_DUMPER	0x04
#define	XID6_DATE	0x05
#define	XID6_EMU	0x06
#define	XID6_CMNTS	0x07
#define	XID6_INTRO	0x30
#define	XID6_LOOP	0x31
#define	XID6_END	0x32
#define	XID6_FADE	0x33
#define	XID6_MUTE	0x34
#define	XID6_LOOPX	0x35
#define	XID6_AMP	0x36
#define	XID6_OST	0x10
#define	XID6_DISC	0x11
#define	XID6_TRACK	0x12
#define	XID6_PUB	0x13
#define	XID6_COPY	0x14

//Data types
#define	XID6_TVAL	0x00
#define	XID6_TSTR	0x01
#define	XID6_TINT	0x04

#define	XID6_EMULIST	8

// compilation variables

const int buffer_duration = 100;			// larger values cause seeking distortion


// foobar2000-configurable variables

static cfg_int cfg_samplerate		("sapu_samplerate",		32000);
static cfg_int cfg_bitspersample	("sapu_bitspersample",	16);
static cfg_int cfg_channels			("sapu_channels",		2);
static cfg_int cfg_mixing			("sapu_mixing",			MIX_INT);
static cfg_int cfg_interpolation	("sapu_interpolation",	INT_GAUSS);
static cfg_int cfg_options			("sapu_options",		0);

static cfg_int cfg_fastseek			("sapu_fastseek",		1);

static cfg_int cfg_usespclength		("sapu_usespclength",	1);
static cfg_int cfg_forcelength		("sapu_forcelength",	1);
static cfg_int cfg_songlength		("sapu_songlength",		180000);
static cfg_int cfg_fadelength		("sapu_fadelength",		2000);
static cfg_int cfg_amplification	("samp_amplification",	16);

// spc file header

struct SPCHEAD
{
	s8	tag[33];							// file tag
	s8	term[3];							// tag terminator
	s8	ver;								// version #/100
	u8	pc[2];								// spc registers
	u8	a;
	u8	x;
	u8	y;
	u8	psw;
	u8	sp;
	u16	__r1;
	s8	song[32];							// song title
	s8	game[32];							// game title
	s8	dumper[16];							// name of dumper
	s8	comment[32];						// comments
	s8	date[11];							// date dumped
	s8	songlen[3];							// song length (seconds before fading)
	s8	fadelen[5];							// fade length (milliseconds)
	s8	artist[32];							// artist
	u8	chandis;							// channels disabled
	u8	emulator;							// emulator used for dump
	u8	__r2[45];
};

struct XID6Chk
{
	u8	id;									// Subchunk ID
	u8	type;								// Type of data
	u16	val;
};

static const s8 emuList[XID6_EMULIST][12] =
		{"unknown","ZSNES","Snes9x","ZST2SPC","ETC","SNEShout","ZSNES/W","Snes9xpp"};

// our input class
class input_snesapu : public input_pcm
{
private:

	SAPUFunc	sapu;						// SNESAPU instance
	HINSTANCE	hDLL;						// SNESAPU dll handle
	static input_snesapu *pDLL_use;
	u32			version, minimum, options;	// SNESAPU version info

	u32			spcfilelen;					// in-memory copy of SPC file
	u8			*spcfile;

	u32			bufferlen;					// output sound buffer
	u8			*buffer;

	u32			sapu_ticks;					// keep track of how many ms have been

	u32			sapu_samplerate;			// internal copies of cfg_* vars
	u32			sapu_bitspersample;
	u32			sapu_channels;
	u32			sapu_mixing;
	u32			sapu_interpolation;
	u32			sapu_options;

	u32			sapu_fastseek;

	u32			sapu_forcelength;
	u32			sapu_usespclength;
	u32			sapu_songlength;
	u32			sapu_fadelength;

	u32			stored_songlength;			// song length info read from spc file
	u32			stored_fadelength;

	// for SSDLabo's snesapu.dll
	u8 *pProgram, *pDSPFlag, *pChange;
	u32 *pWork, *pDetune;

	void try700_(file_info *info)
	{
		if (sapu.InitWork_700) sapu.InitWork_700();

		if (sapu.try700)
		{
			string8 s700(info->get_file_path());
			// file://?
			if (!strncmp(s700, "file://", ::lstrlen("file://"))) {
				int last = s700.find_last("/");
				s700.remove_chars(0, last+1);
				string_ansi_from_utf8 ansi_path(s700);
				sapu.try700((char*)ansi_path.get_ptr());
			}
		}
	}

  public:

	input_snesapu()
		: hDLL(NULL), spcfile(NULL), buffer(NULL)
	{
	}

	bool load_dll()
	{
		if (pDLL_use != NULL) {
			console::info("SNESAPU.DLL does not support multiple instances.");
			console::info("Please stop other decoding processes and try again.");
			return false;
		}

		pDLL_use = this;
		// attempt SNESAPU.dll load
		hDLL = LoadLibrary("SNESAPU.DLL");

		if (hDLL)
		{
			ZeroMemory(&sapu, sizeof(sapu));
			*(void**)&sapu.GetAPUData		= (void*)GetProcAddress(hDLL, "GetAPUData");
			*(void**)&sapu.SNESAPUInfo		= (void*)GetProcAddress(hDLL, "SNESAPUInfo");
			*(void**)&sapu.ResetAPU			= (void*)GetProcAddress(hDLL, "ResetAPU");
			*(void**)&sapu.FixAPU			= (void*)GetProcAddress(hDLL, "FixAPU");
			*(void**)&sapu.LoadSPCFile		= (void*)GetProcAddress(hDLL, "LoadSPCFile");
			*(void**)&sapu.EmuAPU			= (void*)GetProcAddress(hDLL, "EmuAPU");
			*(void**)&sapu.EmuDSP			= (void*)GetProcAddress(hDLL, "EmuDSP");
			*(void**)&sapu.EmuSPC			= (void*)GetProcAddress(hDLL, "EmuSPC");
			*(void**)&sapu.SetAPUOpt		= (void*)GetProcAddress(hDLL, "SetAPUOpt");
			*(void**)&sapu.SetAPULength		= (void*)GetProcAddress(hDLL, "SetAPULength");
			*(void**)&sapu.SetDSPAmp		= (void*)GetProcAddress(hDLL, "SetDSPAmp");
			*(void**)&sapu.SetDSPReg		= (void*)GetProcAddress(hDLL, "SetDSPReg");
			*(void**)&sapu.SeekAPU			= (void*)GetProcAddress(hDLL, "SeekAPU");

			*(void**)&sapu.SetTimerTrick	= (void*)GetProcAddress(hDLL, "SetTimerTrick");
			*(void**)&sapu.GetScript700Data	= (void*)GetProcAddress(hDLL, "GetScript700Data");
			*(void**)&sapu.SetScript700		= (void*)GetProcAddress(hDLL, "SetScript700");
			*(void**)&sapu.try700		= (void*)GetProcAddress(hDLL,"try700");
			*(void**)&sapu.InitWork_700	= (void*)GetProcAddress(hDLL,"InitWork_700");

			sapu.GetAPUData(&sapu.ram, &sapu.xram, NULL, &sapu.t64Cnt,
							&sapu.dsp, &sapu.voice, &sapu.vMMaxL, &sapu.vMMaxR);

			if (!sapu.SetScript700 && sapu.GetScript700Data) {
				sapu.GetScript700Data(NULL, &pProgram, &pWork, NULL, NULL, NULL, NULL, &pDSPFlag, &pChange, &pDetune);
			}

			if (sapu.SNESAPUInfo == NULL || sapu.LoadSPCFile == NULL || sapu.EmuAPU == NULL)
			{
				console::info("SNESAPU.DLL function lookups failed.");
			}
			else
			{
				sapu.SNESAPUInfo(&version, &minimum, &options);

				if (version < 0x20000 || minimum > 0x20000)
				{
					console::info("Incompatible SNESAPU.DLL version.");
					hDLL = 0;
					return false;
				}
			}
		}
		else
		{
			console::info("Could not locate SNESAPU.DLL.");
			return false;
		}
		return true;
	}

	~input_snesapu()
	{
		if (hDLL)
		{
			FreeLibrary(hDLL);
			pDLL_use = NULL;
		}

		if (spcfile)
		{
			delete [] spcfile;
		}

		if (buffer)
		{
			delete [] buffer;
		}
	}

	virtual bool test_filename(const char *name, const char *ext)
	{
		return (stricmp_utf8(ext,"SPC") == 0);
	}

	int spc_header_num_test(char *str, int len)
	{
		int	pos = 0;

		while ((pos < len) && ((str[pos] >= 0x30 && str[pos] <= 0x39) || str[pos] == '/'))
		{
			++pos;
		}

		if (pos == len || str[pos] == 0)
		{
			return pos;
		}
		else
		{
			return -1;
		}
	}

	bool spc_header_read(SPCHEAD *spchead, file_info *info, unsigned flags)
	{
		// Check header
		if (memcmp(spchead, SPC_ID, 28)) return false;

		string8 spc_song;
		string8 spc_game;
		string8 spc_artist;
		string8 spc_dumper;
		string8 spc_comment;
		string8 spc_ost;
		string8 spc_ost_disc;
		string8 spc_ost_track;
		char spc_songlength[4];
		char spc_fadelength[6];
		string8 spc_publisher;
		string8 spc_copylight;
		string8 spc_emu;

		// initialise temp strings
		spc_songlength[32] = 0;
		spc_fadelength[32] = 0;

		spc_song.set_string_ansi(spchead->song, 32);
		spc_game.set_string_ansi(spchead->game, 32);
		spc_artist.set_string_ansi(spchead->artist, 32);
		spc_dumper.set_string_ansi(spchead->dumper, 16);
		spc_comment.set_string_ansi(spchead->comment, 32);
		memcpy(spc_songlength, spchead->songlen, 3);
		memcpy(spc_fadelength, spchead->fadelen, 5);
		if (spchead->emulator >= 0x30 && spchead->emulator < 0x30+XID6_EMULIST)
		{
			spc_emu.set_string(emuList[spchead->emulator - 0x30]);
		}

		stored_songlength = atoi(spc_songlength) * 1000;
		stored_fadelength = atoi(spc_fadelength);

		// clamp song lengths
		if (stored_songlength > 959000)
		{
			stored_songlength = 959000;
		}
		if (stored_fadelength > 59999)
		{
			stored_fadelength = 59999;
		}

		// decide song length
		if (stored_songlength && sapu_usespclength)
		{
			sapu_forcelength = 1;
			sapu_songlength  = stored_songlength;
			sapu_fadelength  = stored_fadelength;
		}

		// Extended ID666
		if (spcfilelen >= 0x1020c) {
			if (memcmp(&spcfile[OFS_XID6], XID6_ID, 4) == 0) {
				s32 hdr_size = *((u32*)&spcfile[OFS_XID6 + 4]);
				s32 remain_size = spcfilelen - (OFS_XID6+8);
				if (hdr_size > remain_size) hdr_size = remain_size;

				XID6Chk *id6 = (XID6Chk*)&spcfile[OFS_XID6+8];
				while (hdr_size > 0) {
					hdr_size -= 4;
					int length = (id6->val + 3) & ~3;
					switch (id6->id) {
						case XID6_SONG:
							spc_song.set_string_ansi((const char*)((&id6->val)+1), id6->val);
							break;
						case XID6_GAME:
							spc_game.set_string_ansi((const char*)((&id6->val)+1), id6->val);
							break;
						case XID6_ARTIST:
							spc_artist.set_string_ansi((const char*)((&id6->val)+1), id6->val);
							break;
						case XID6_PUB:
							spc_publisher.set_string_ansi((const char*)((&id6->val)+1), id6->val);
							break;
						case XID6_OST:
							spc_ost.set_string_ansi((const char*)((&id6->val)+1), id6->val);
							break;
						case XID6_DUMPER:
							spc_dumper.set_string_ansi((const char*)((&id6->val)+1), id6->val);
							break;
						case XID6_CMNTS:
							spc_comment.set_string_ansi((const char*)((&id6->val)+1), id6->val);
							break;
						case XID6_DISC:
							spc_ost_disc.add_int(id6->val);
							break;
						case XID6_TRACK:
							spc_ost_track.add_int(id6->val >> 8);
							break;
						case XID6_COPY:
							spc_copylight.add_int(id6->val);
							break;
						case XID6_EMU:
							if (id6->val >= 0x30 && id6->val < 0x30+XID6_EMULIST) {
								spc_emu.add_string(emuList[id6->val - 0x30]);
							}
							break;
						// Not implement yet.
						case XID6_MUTE:
							break;
						case XID6_LOOPX:
							break;
						case XID6_AMP:
							break;
						case XID6_DATE:
							break;
						case XID6_FADE:
							break;
						case XID6_INTRO:
							break;
						case XID6_LOOP:
							break;
						case XID6_END:
							break;

					}
					if (id6->type != XID6_TVAL) {
						id6 = (XID6Chk*)((s32)id6 + length);
						hdr_size -= length;
					}
					id6++;
				}
			}
		}

		// send info to foobar2000
		if (flags & OPEN_FLAG_GET_INFO) {
			if (sapu_forcelength) {
				info->info_set_int("spc_songlength", stored_songlength / 1000);
				info->info_set_int("spc_fadelength", stored_fadelength);
				info->set_length((double)(sapu_songlength + sapu_fadelength) / 1000.0);
			}
			if (spc_song.is_empty() == false)
			{
				info->meta_set("TITLE", spc_song);
			}
			if (spc_game.is_empty() == false)
			{
				info->meta_set("ALBUM", spc_game);
			}
			if (spc_artist.is_empty() == false)
			{
				info->meta_set("ARTIST", spc_artist);
			}
			if (spc_dumper.is_empty() == false)
			{
				info->meta_set("DUMPER", spc_dumper);
			}
			if (spc_comment.is_empty() == false)
			{
				info->meta_set("COMMENT", spc_comment);
			}
			if (spc_publisher.is_empty() == false) {
				info->meta_set("PUBLISHER", spc_publisher);
			}
			if (spc_copylight.is_empty() == false) {
				info->meta_set("COPYRIGHT", spc_copylight);
			}
			if (spc_ost.is_empty() == false) {
				info->meta_set("OST", spc_ost);
			}
			if (spc_ost_disc.is_empty() == false) {
				info->meta_set("OST_DISC", spc_ost_disc);
			}
			if (spc_ost_track.is_empty() == false) {
				info->meta_set("OST_TRACK", spc_ost_track);
			}
			if (spc_emu.is_empty() == false) {
				info->meta_set("EMULATOR", spc_emu);
			}

			info->info_set_int("samplerate", sapu_samplerate);
			info->info_set_int("channels", sapu_channels);
			info->info_set_int("bitspersample", sapu_bitspersample);
		}

		return true;
	}

	bool open(reader *r, file_info *info, unsigned flags)
	{
		// read spc file
		spcfilelen			= (u32) r->get_length();
		spcfile				= new u8[spcfilelen];
		if (r->read(spcfile, spcfilelen) == 0) return false;

		// load settings
		sapu_samplerate		= cfg_samplerate;
		sapu_bitspersample	= cfg_bitspersample;
		sapu_channels		= cfg_channels;
		sapu_mixing			= cfg_mixing;
		sapu_interpolation	= cfg_interpolation;
		sapu_options		= cfg_options;

		sapu_fastseek		= cfg_fastseek;

		sapu_forcelength	= cfg_forcelength;
		sapu_usespclength	= cfg_usespclength;
		sapu_songlength		= cfg_songlength;
		sapu_fadelength		= cfg_fadelength;

		// read info from spc
		if (spc_header_read((SPCHEAD*) spcfile, info, flags) == false) return false;

		if (flags & OPEN_FLAG_DECODE)
		{
			// load SNESAPU.DLL
			if (load_dll() == false) return false;

			// send options/file to snesapu
			sapu.LoadSPCFile(spcfile);
			sapu.SetAPUOpt(sapu_mixing, sapu_channels, sapu_bitspersample, sapu_samplerate, sapu_interpolation, sapu_options);
			if (sapu.GetScript700Data)
			{
				if (sapu.SetScript700)
				{
					script700_parser parser;
					if (parser.load(info) == true)
					{
						void *pScript = parser.get_ptr();
						if (pScript)
						{
							if (sapu.SetScript700(pScript) == -1)
							{
								console::info("Script700 parse error");
							}
						}
					}
				}
				else
				{
					script700_parser parser(pProgram, pWork, pDSPFlag, pChange, pDetune);
					if (parser.load(info) == true) {
						parser.parse();
					}
				}
			}
			else
			{
				try700_(info);
			}

			// allocate our output buffer
			bufferlen	= (u32)(((buffer_duration * sapu_samplerate) + buffer_duration) / 1000);
			bufferlen	= bufferlen * sapu_channels * (sapu_bitspersample / 8);
			buffer		= new u8[bufferlen];

			sapu.SetDSPAmp(cfg_amplification);

			sapu_ticks = 0;
			if (sapu_forcelength)
			{
				sapu.SetAPULength(sapu_songlength * 64, sapu_fadelength * 64);
			}
		}

		// if finite playtime is wanted, enable forced length
		if (flags & OPEN_FLAG_NO_LOOPING)
		{
			sapu_usespclength = sapu_forcelength = 1;
		}

		return 1;
	}

	virtual int get_samples_pcm(void **out_buffer, int *out_size, int *srate, int *bps, int *nch)
	{
		// decide how many samples we want from sapu
		int mswanted = buffer_duration;

		// clip samples requested if nearing song end
		if (sapu_forcelength)
		{
			if (sapu_ticks >= (sapu_songlength + sapu_fadelength))
			{
				return 0;
			}

			if ( (sapu_ticks + mswanted) > (sapu_songlength + sapu_fadelength) )
			{
				mswanted = (sapu_songlength + sapu_fadelength) - sapu_ticks;
			}
		}

		//EmuAPU() wants no. of samples to render
		sapu.EmuAPU(buffer, ((mswanted * sapu_samplerate) / 1000), true);

		*out_buffer = buffer;
		*out_size = ((mswanted * sapu_samplerate) / 1000) * sapu_channels * (sapu_bitspersample / 8);
		*srate = sapu_samplerate;
		*bps = sapu_bitspersample;
		*nch = sapu_channels;

		sapu_ticks += mswanted;

		return 1;
	}

    virtual set_info_t set_info(reader *r, const file_info *info)
	{
		return SET_INFO_FAILURE;
	}

	virtual bool seek(double seconds)
	{
		// compare desired seek pos with current pos
		u32 newpos = (u32)(seconds * 1000.0);

		// reset apu if necessary
		if (*sapu.t64Cnt >= newpos<<6)
		{
			sapu.LoadSPCFile(spcfile);
		}
		sapu.SeekAPU(((newpos<<6) - *sapu.t64Cnt), sapu_fastseek);

		sapu_ticks = newpos;

		return 1;
	}

	virtual bool can_seek()
	{
		// we can only seek if the song has a length
		if (sapu_forcelength)
		{
			return 1;
		}
		else
		{
			return 0;
		}
	}
};

static service_factory_t<input,input_snesapu> foo;


// combo populator helper

void combo_populate(HWND hwnd, int combo, const char **names, int *vals, int len, int def)
{
	int i;
	int index;
	int defindex = 0;

	uSendDlgItemMessage(hwnd, combo, CB_RESETCONTENT, 0, 0);

	for (i=0 ; i<len ; ++i)
	{
		index = uSendDlgItemMessageText(hwnd, combo, CB_INSERTSTRING, -1, names[i]);
		uSendDlgItemMessage(hwnd, combo, CB_SETITEMDATA, index, vals[i]);

		if (vals[i] == def)
		{
			defindex = i;
		}
	}

	uSendDlgItemMessage(hwnd, combo, CB_SETCURSEL, defindex, -1);
}


// combobox setting helper

#define COMBOGET(X) \
	if (HIWORD(wparam) == CBN_SELCHANGE) \
	{ \
		data = uSendMessage((HWND)lparam, CB_GETCURSEL, 0, 0); \
		data = uSendMessage((HWND)lparam, CB_GETITEMDATA, data, 0); \
		if (data != CB_ERR) \
		{ \
			X = data; \
		} \
	} \


// checkbox setting helpers

#define DSPCHECKGET(X) \
	if (uSendMessage((HWND)lparam, BM_GETCHECK, 0, 0) == BST_CHECKED) \
	{ \
		cfg_options = cfg_options | X; \
	} \
	else if (cfg_options & X) \
	{ \
		cfg_options = cfg_options - X; \
	} \

#define CHECKGET(X) \
	if (uSendMessage((HWND)lparam, BM_GETCHECK, 0, 0) == BST_CHECKED) \
	{ \
		X = 1; \
	} \
	else \
	{ \
		X = 0; \
	} \


// configuration window procedure

static BOOL CALLBACK ConfigProc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
	int data;

	switch (msg)
	{
		case WM_INITDIALOG:
		{
			// populate comboboxes
			const char *sampleratenames[] = {"8000","11025","16000","22050","24000","32000","44100","48000","64000","88200","96000"};
			int samplerates[] =				{ 8000,  11025,  16000,  22050,  24000,  32000,  44100,  48000,  64000,  88200,  96000};

			const char *mixingnames[] =		{"None (faulty?)","Integer","Float"};
			int mixingmodes[] =				{ 0,	 1,		   2};

			const char *internames[] =		{"None","Linear","Cubic","Gaussian"};
			int intermodes[] =				{ 0,	 1,		  2,	  3};

			const char *bitnames[] =		{"8","16","24","32"};
			int bitmodes[] =				{ 8,  16,  24,  32};

			const char *channames[] =		{"Mono","Stereo"};
			int chanmodes[] =				{ 1,	 2};

			const char *trickportmes[] =	{"0","1","2","3"};
			int trickportmodes[] =			{0,1,2,3};

			const char *tricktimemes[] =	{"0","1","2","147","483","647"};
			int tricktimemodes[] =			{0,1,2,147,483,647};

			combo_populate(hwnd, IDC_SAMPLERATE,	sampleratenames, samplerates, 11, cfg_samplerate);
			combo_populate(hwnd, IDC_MIXING,		mixingnames,	 mixingmodes, 3,  cfg_mixing);
			combo_populate(hwnd, IDC_INTERPOLATION, internames,		 intermodes,  4,  cfg_interpolation);
			combo_populate(hwnd, IDC_BITS,			bitnames,		 bitmodes,	  4,  cfg_bitspersample);
			combo_populate(hwnd, IDC_CHANNELS,		channames,		 chanmodes,   2,  cfg_channels);

			// populate dsp options
			int dspopts[][2] = {
				{IDC_DSP_ANALOG,	DSP_ANALOG},
				{IDC_DSP_OLDADPCM,	DSP_OLDSMP},
				{IDC_DSP_SURROUND,	DSP_SURND},
				{IDC_DSP_REVERSE,	DSP_REVERSE},
				{IDC_DSP_DISECHO,	DSP_NOECHO}};

			for (int i=0 ; i<5 ; ++i)
			{
				uSendDlgItemMessage(hwnd, dspopts[i][0], BM_SETCHECK, cfg_options & dspopts[i][1], 0);
			}

			// populate song length settings
			char tmp[80];

			uSendDlgItemMessage(hwnd, IDC_FORCELENGTH,  BM_SETCHECK, cfg_forcelength, 0);
			uSendDlgItemMessage(hwnd, IDC_USESPCLENGTH,  BM_SETCHECK, cfg_usespclength, 0);

			data = cfg_songlength;
			wsprintf(tmp, "%d", data);
			uSetDlgItemText(hwnd, IDC_SONGLENGTH, (const char*) tmp);

			data = cfg_fadelength;
			wsprintf(tmp, "%d", data);
			uSetDlgItemText(hwnd, IDC_FADELENGTH, (const char*) tmp);

			// populate misc options
			uSendDlgItemMessage(hwnd, IDC_FASTSEEK,  BM_SETCHECK, cfg_fastseek, 0);

			HWND hSlider = GetDlgItem(hwnd, IDC_VOLUME);
			::uSendMessage(hSlider, TBM_SETTICFREQ, 16, 0) ;
			::uSendMessage(hSlider, TBM_SETRANGE, (WPARAM)TRUE, (LPARAM)MAKELONG(0, 255));
			::uSendMessage(hSlider, TBM_SETPOS, (WPARAM)TRUE, (LPARAM)(cfg_amplification));
			string8 volume;
			volume.add_int(cfg_amplification);
			::SetDlgItemText(hwnd, IDC_VOLUME_VALUE, volume);

			return 1;
		}

		case WM_COMMAND:
		{
			switch (LOWORD(wparam))
			{
				case IDC_SAMPLERATE:
				{
					COMBOGET(cfg_samplerate)
					break;
				}

				case IDC_MIXING:
				{
					COMBOGET(cfg_mixing)
					break;
				}

				case IDC_INTERPOLATION:
				{
					COMBOGET(cfg_interpolation)
					break;
				}

				case IDC_BITS:
				{
					COMBOGET(cfg_bitspersample)
					break;
				}

				case IDC_CHANNELS:
				{
					COMBOGET(cfg_channels)
					break;
				}

				case IDC_DSP_ANALOG:
				{
					DSPCHECKGET(DSP_ANALOG)
					break;
				}

				case IDC_DSP_OLDADPCM:
				{
					DSPCHECKGET(DSP_OLDSMP)
					break;
				}

				case IDC_DSP_SURROUND:
				{
					DSPCHECKGET(DSP_SURND)
					break;
				}

				case IDC_DSP_REVERSE:
				{
					DSPCHECKGET(DSP_REVERSE)
					break;
				}

				case IDC_DSP_DISECHO:
				{
					DSPCHECKGET(DSP_NOECHO)
					break;
				}

				case IDC_USESPCLENGTH:
				{
					CHECKGET(cfg_usespclength)
					break;
				}

				case IDC_FORCELENGTH:
				{
					CHECKGET(cfg_forcelength)
					break;
				}

				case IDC_SONGLENGTH:
				{
					if (HIWORD(wparam) == EN_CHANGE)
					{
						data = atoi(string_utf8_from_window((HWND)lparam));

						if (data > 0)
						{
							cfg_songlength = data;
						}
					}
					break;
				}

				case IDC_FADELENGTH:
				{
					if (HIWORD(wparam) == EN_CHANGE)
					{
						data = atoi(string_utf8_from_window((HWND)lparam));

						if (data > 0)
						{
							cfg_fadelength = data;
						}
					}
					break;
				}

				case IDC_FASTSEEK:
				{
					CHECKGET(cfg_fastseek)
					break;
				}
			}
		}
		case WM_HSCROLL:
		{
			switch (GetWindowLong((HWND)lparam, GWL_ID))
			{
				case IDC_VOLUME:
				{
					cfg_amplification = (int)SendMessage((HWND)lparam, TBM_GETPOS, 0, 0);
					string8 value;
					value.add_int(cfg_amplification);
					::SetDlgItemText(hwnd, IDC_VOLUME_VALUE, value);
				}
			}
		}
	}

	return 0;
}


// our config class

class config_snesapu : public config
{
public:

	virtual HWND create(HWND parent)
	{
		return uCreateDialog(IDD_CONFIG, parent, ConfigProc);
	}

	virtual const char *get_name()
	{
		return "SNESAPU SPC Decoder";
	}

	virtual const char *get_parent_name()
	{
		return "Input";
	}
};

static service_factory_single_t<config,config_snesapu> foo_snesapu_config;


// version info

DECLARE_COMPONENT_VERSION("SNESAPU SPC Decoder",
						  "0.01a",
						  "Uses SNESAPU.dll for SPC decoding.")

DECLARE_FILE_TYPE("SPC files","*.spc" );

input_snesapu * input_snesapu::pDLL_use = NULL;
